@@ -15,8 +15,7 @@ import substra
 import substratools
 
 import substrafl
-from substrafl import exceptions
-from substrafl.algorithms.algo import Algo
+from substrafl.compute_plan_builder import ComputePlanBuilder
 from substrafl.dependency import Dependency
 from substrafl.evaluation_strategy import EvaluationStrategy
 from substrafl.exceptions import KeyMetadataError
@@ -25,7 +24,6 @@ from substrafl.nodes.aggregation_node import AggregationNode
 from substrafl.nodes.node import OperationKey
 from substrafl.nodes.train_data_node import TrainDataNode
 from substrafl.remote.remote_struct import RemoteStruct
-from substrafl.strategies.strategy import Strategy
 
 logger = logging.getLogger(__name__)
 
@@ -37,7 +35,7 @@ def _register_operations(
     evaluation_strategy: Optional[EvaluationStrategy],
     dependencies: Dependency,
 ) -> Tuple[List[dict], Dict[RemoteStruct, OperationKey]]:
-    """Register the operations in Substra: define the algorithms we need and submit them
+    """Register the operations in Substra: define the functions we need and submit them
 
     Args:
         client (substra.Client): substra client
@@ -46,57 +44,70 @@ def _register_operations(
             centralized strategies
         evaluation_strategy (typing.Optional[EvaluationStrategy]): the evaluation strategy
             if there is one dependencies
-        (Dependency): dependencies of the train algo
+        (Dependency): dependencies of the experiment
 
     Returns:
         typing.Tuple[typing.List[dict], typing.Dict[RemoteStruct, OperationKey]]:
         tasks, operation_cache
     """
     # `register_operations` methods from the different organizations store the id of the already registered
-    # algorithm so we don't add them twice
+    # functions so we don't add them twice
     operation_cache = dict()
-    predict_algo_cache = dict()
+    predict_function_cache = dict()
+    test_function_cache = dict()
     tasks = list()
 
     train_data_organizations_id = {train_data_node.organization_id for train_data_node in train_data_nodes}
     aggregation_organization_id = {aggregation_node.organization_id} if aggregation_node is not None else set()
+    test_data_organizations_ids = (
+        evaluation_strategy.test_data_nodes_org_ids if evaluation_strategy is not None else set()
+    )
 
-    authorized_ids = list(train_data_organizations_id | aggregation_organization_id)
+    authorized_ids = list(train_data_organizations_id | aggregation_organization_id | test_data_organizations_ids)
     permissions = substra.sdk.schemas.Permissions(public=False, authorized_ids=authorized_ids)
 
     for train_data_node in train_data_nodes:
         operation_cache = train_data_node.register_operations(
-            client,
-            permissions,
+            client=client,
+            permissions=permissions,
             cache=operation_cache,
             dependencies=dependencies,
         )
 
-        tasks += train_data_node.tuples
+        if train_data_node.init_task is not None:
+            tasks += [train_data_node.init_task]
+
+        tasks += train_data_node.tasks
 
     if evaluation_strategy is not None:
         for test_data_node in evaluation_strategy.test_data_nodes:
-            predict_algo_cache = test_data_node.register_predict_operations(
-                client,
-                permissions,
-                cache=predict_algo_cache,
+            predict_function_cache = test_data_node.register_predict_operations(
+                client=client,
+                permissions=permissions,
+                cache=predict_function_cache,
+                dependencies=dependencies,
+            )
+            test_function_cache = test_data_node.register_test_operations(
+                client=client,
+                permissions=permissions,
+                cache=test_function_cache,
                 dependencies=dependencies,
             )
 
-            tasks += test_data_node.predicttuples
-            tasks += test_data_node.testtuples
+            tasks += test_data_node.predicttasks
+            tasks += test_data_node.testtasks
 
     # The aggregation operation is defined in the strategy, its dependencies are
     # the strategy dependencies
     if aggregation_node is not None:
         operation_cache = aggregation_node.register_operations(
-            client,
-            permissions,
+            client=client,
+            permissions=permissions,
             cache=operation_cache,
-            dependencies=Dependency(editable_mode=dependencies.editable_mode),
+            dependencies=dependencies,
         )
 
-        tasks += aggregation_node.tuples
+        tasks += aggregation_node.tasks
 
     return tasks, operation_cache
 
@@ -104,11 +115,10 @@ def _register_operations(
 def _save_experiment_summary(
     experiment_folder: Path,
     compute_plan_key: str,
-    strategy: Strategy,
+    strategy: ComputePlanBuilder,
     num_rounds: int,
-    algo: Algo,
     operation_cache: Dict[RemoteStruct, OperationKey],
-    train_data_nodes: TrainDataNode,
+    train_data_nodes: List[TrainDataNode],
     aggregation_node: Optional[AggregationNode],
     evaluation_strategy: EvaluationStrategy,
     timestamp: str,
@@ -121,9 +131,8 @@ def _save_experiment_summary(
         compute_plan_key (str): compute plan key
         strategy (substrafl.strategies.Strategy): strategy
         num_rounds (int): num_rounds
-        algo (substrafl.algorithms.Algo): algo
         operation_cache (typing.Dict[RemoteStruct, OperationKey]): operation_cache
-        train_data_nodes (TrainDataNode): train_data_nodes
+        train_data_nodes (typing.List[TrainDataNode]): train_data_nodes
         aggregation_node (typing.Optional[AggregationNode]): aggregation_node
         evaluation_strategy (EvaluationStrategy): evaluation_strategy
         timestamp (str): timestamp with "%Y_%m_%d_%H_%M_%S" format
@@ -137,9 +146,9 @@ def _save_experiment_summary(
     # add attributes of interest and summaries of the classes to the experiment summary
     experiment_summary["compute_plan_key"] = compute_plan_key
     experiment_summary["strategy"] = type(strategy).__name__
-    experiment_summary["num_rounds"] = num_rounds
-    experiment_summary["algo"] = algo.summary()
-    experiment_summary["algo_keys"] = {
+    if num_rounds is not None:
+        experiment_summary["num_rounds"] = num_rounds
+    experiment_summary["function_keys"] = {
         operation_cache[remote_struct]: remote_struct.summary() for remote_struct in operation_cache
     }
     experiment_summary["train_data_nodes"] = [train_data_node.summary() for train_data_node in train_data_nodes]
@@ -173,9 +182,7 @@ def _check_evaluation_strategy(
 
 
 def _check_additional_metadata(additional_metadata: Dict):
-    unauthorized_keys = set(
-        ("substrafl_version", "substra_version", "substratools_version", "python_version", "num_rounds")
-    )
+    unauthorized_keys = {"substrafl_version", "substra_version", "substratools_version", "python_version"}
     invalid_keys = set(additional_metadata.keys()).intersection(unauthorized_keys)
 
     if len(invalid_keys) > 0:
@@ -207,12 +214,12 @@ def _get_packages_versions() -> dict:
 
 
 def execute_experiment(
+    *,
     client: substra.Client,
-    algo: Algo,
-    strategy: Strategy,
+    strategy: ComputePlanBuilder,
     train_data_nodes: List[TrainDataNode],
-    num_rounds: int,
     experiment_folder: Union[str, Path],
+    num_rounds: Optional[int] = None,
     aggregation_node: Optional[AggregationNode] = None,
     evaluation_strategy: Optional[EvaluationStrategy] = None,
     dependencies: Optional[Dependency] = None,
@@ -222,10 +229,10 @@ def execute_experiment(
     task_submission_batch_size: int = 500,
 ) -> substra.sdk.models.ComputePlan:
     """Run a complete experiment. This will train (on the `train_data_nodes`) and test (on the
-    `test_data_nodes`) your `algo` with the specified `strategy` `n_rounds` times and return the
+    `test_data_nodes`) the specified `strategy` `n_rounds` times and return the
     compute plan object from the Substra platform.
 
-    In substrafl, operations are linked to each other statically before being submitted to substra.
+    In SubstraFL, operations are linked to each other statically before being submitted to Substra.
 
     The execution of:
 
@@ -234,7 +241,7 @@ def execute_experiment(
 
     generate the static graph of operations.
 
-    Each element necessary for those operations (Tasks and Algorithms)
+    Each element necessary for those operations (Tasks and Functions)
     is registered to the Substra platform thanks to the specified client.
 
     Finally, the compute plan is sent and executed.
@@ -243,16 +250,15 @@ def execute_experiment(
 
     Args:
         client (substra.Client): A substra client to interact with the Substra platform
-        algo (Algo): The algorithm your strategy will execute (i.e. train and test on all the specified nodes)
-        strategy (Strategy): The strategy by which your algorithm will be executed
+        strategy (Strategy): The strategy that will be executed
         train_data_nodes (typing.List[TrainDataNode]): List of the nodes where training on data
             occurs evaluation_strategy (EvaluationStrategy, Optional): If None performance will not be measured at all.
             Otherwise measuring of performance will follow the EvaluationStrategy. Defaults to None.
         aggregation_node (typing.Optional[AggregationNode]): For centralized strategy, the aggregation
             node, where all the shared tasks occurs else None.
         num_rounds (int): The number of time your strategy will be executed
-        dependencies (Dependency, Optional): Dependencies of the algorithm. It must be defined from
-            the substrafl Dependency class. Defaults None.
+        dependencies (Dependency, Optional): Dependencies of the experiment. It must be defined from
+            the SubstraFL Dependency class. Defaults None.
         experiment_folder (typing.Union[str, pathlib.Path]): path to the folder where the experiment summary is saved.
         clean_models (bool): Clean the intermediary models on the Substra platform. Set it to False
             if you want to download or re-use intermediary models. This causes the disk space to fill
@@ -270,13 +276,6 @@ def execute_experiment(
     if dependencies is None:
         dependencies = Dependency()
 
-    if strategy.name not in algo.strategies:
-        raise exceptions.IncompatibleAlgoStrategyError(
-            f"The algo {algo.__class__.__name__} is not compatible with the strategy {strategy.__class__.__name__},"
-            f"named {strategy.name}. Check the algo strategies property: algo.strategies to see the list of compatible"
-            "strategies."
-        )
-
     train_data_nodes = copy.deepcopy(train_data_nodes)
     aggregation_node = copy.deepcopy(aggregation_node)
     strategy = copy.deepcopy(strategy)
@@ -285,7 +284,7 @@ def execute_experiment(
     train_organization_ids = [train_data_node.organization_id for train_data_node in train_data_nodes]
 
     if len(train_organization_ids) != len(set(train_organization_ids)):
-        raise ValueError("Training multiple algorithms on the same organization is not supported right now.")
+        raise ValueError("Training multiple functions on the same organization is not supported right now.")
 
     if evaluation_strategy is not None:
         _check_evaluation_strategy(evaluation_strategy, num_rounds)
@@ -300,31 +299,18 @@ def execute_experiment(
     # Adding substrafl, substratools and substra versions to the cp metadata
     cp_metadata.update(_get_packages_versions())
 
-    # Adding rounds metadata to cp_meta_data
-    cp_metadata["num_rounds"] = num_rounds
-
     logger.info("Building the compute plan.")
 
-    # create computation graph
-    for round_idx in range(0, num_rounds + 1):
-        strategy.perform_round(
-            algo=algo,
-            train_data_nodes=train_data_nodes,
-            aggregation_node=aggregation_node,
-            round_idx=round_idx,
-            clean_models=clean_models,
-        )
-
-        if evaluation_strategy is not None and next(evaluation_strategy):
-            strategy.predict(
-                algo=algo,
-                train_data_nodes=train_data_nodes,
-                test_data_nodes=evaluation_strategy.test_data_nodes,
-                round_idx=round_idx,
-            )
+    strategy.build_compute_plan(
+        train_data_nodes=train_data_nodes,
+        aggregation_node=aggregation_node,
+        evaluation_strategy=evaluation_strategy,
+        num_rounds=num_rounds,
+        clean_models=clean_models,
+    )
 
     # Computation graph is created
-    logger.info("Registering the algorithm to Substra.")
+    logger.info("Registering the functions to Substra.")
     tasks, operation_cache = _register_operations(
         client=client,
         train_data_nodes=train_data_nodes,
@@ -346,7 +332,6 @@ def execute_experiment(
         compute_plan_key=compute_plan_key,
         strategy=strategy,
         num_rounds=num_rounds,
-        algo=algo,
         operation_cache=operation_cache,
         train_data_nodes=train_data_nodes,
         aggregation_node=aggregation_node,
