@@ -1,10 +1,8 @@
 """
-Create the Substra algo assets and register them to the platform.
+Create the Substra function assets and register them to the platform.
 """
-import inspect
 import logging
 import os
-import shutil
 import tarfile
 import tempfile
 import typing
@@ -19,21 +17,19 @@ from packaging import version
 
 import substrafl
 from substrafl import exceptions
+from substrafl.constants import TMP_SUBSTRAFL_PREFIX
 from substrafl.dependency import Dependency
 from substrafl.nodes.node import InputIdentifiers
-from substrafl.nodes.node import OutputIdentifiers
 from substrafl.remote.register.generate_wheel import local_lib_wheels
 from substrafl.remote.register.generate_wheel import pypi_lib_wheels
 from substrafl.remote.remote_struct import RemoteStruct
-from substrafl.remote.substratools_methods import RemoteDataMethod
+from substrafl.remote.substratools_methods import RemoteMethod
 
 logger = logging.getLogger(__name__)
 
 # Substra tools version for which the image naming scheme changed
 MINIMAL_DOCKER_SUBSTRATOOLS_VERSION = "0.16.0"
 
-TMP_SUBSTRAFL_PREFIX = "tmp_substrafl"
-
 _DEFAULT_SUBSTRATOOLS_IMAGE = "ghcr.io/substra/substra-tools:\
 {substratools_version}-nvidiacuda11.8.0-base-ubuntu22.04-python{python_version}"
 
@@ -42,8 +38,6 @@ SUBSTRAFL_FOLDER = "substrafl_internal"
 DOCKERFILE_TEMPLATE = """
 FROM {docker_image}
 
-COPY . .
-
 # install dependencies
 RUN python{python_version} -m pip install -U pip
 
@@ -56,10 +50,13 @@ RUN python{python_version} -m pip install -U pip
 # Install local dependencies
 {local_dependencies}
 
-ENTRYPOINT ["python{python_version}", "algo.py", "--function-name", "{method_name}"]
+ENTRYPOINT ["python{python_version}", "function.py", "--function-name", "{method_name}"]
+
+COPY . .
+
 """
 
-ALGO = """
+FUNCTION = """
 import json
 import cloudpickle
 
@@ -73,61 +70,34 @@ if __name__ == "__main__":
     # Load the wrapped user code
     remote_struct = RemoteStruct.load(src=Path(__file__).parent / '{substrafl_folder}')
 
-    # Create a Substra algo from the wrapped user code
+    # Create a Substra function from the wrapped user code
     remote_instance = remote_struct.get_remote_instance()
 
     # Register the functions to substra-tools
-    remote_instance.register_substratools_functions()
+    remote_instance.register_substratools_function()
 
-    # Execute the algo using substra-tools
+    # Execute the function using substra-tools
     tools.execute()
 """
 
 
-def _copy_local_packages(
-    path: Path, local_dependencies: typing.List[Path], python_major_minor: str, operation_dir: Path
-):
+def iter_dockerfile_install_command(paths: typing.List[Path], python_major_minor: str) -> str:
+    for path in paths:
+        # Does not work with path.is_dir(), surely a race condition. Should be removed in  #123
+        formatted_path = (str(path) + "/") if not path.is_file() else path
+        yield f"COPY {path}  {path} \nRUN python{python_major_minor} -m pip install --no-cache-dir {formatted_path} \n"
+
+
+def _copy_local_packages(path: Path, python_major_minor: str, operation_dir: Path, dependencies: Dependency):
     """Copy the local libraries given by the user and generate the installation command."""
-    dependencies_buffer = list()
     path.mkdir(exist_ok=True)
-    for dependency_path in local_dependencies:
-        dest_path = path / dependency_path.name
-        if dependency_path.is_dir():
-            shutil.copytree(
-                dependency_path,
-                dest_path,
-                ignore=shutil.ignore_patterns(
-                    "local-worker",
-                    TMP_SUBSTRAFL_PREFIX + "*",
-                ),
-            )
-        elif dependency_path.is_file():
-            shutil.copy(dependency_path, dest_path)
-        else:
-            raise ValueError(f"Does not exist {dependency_path}")
+    dependencies_paths = dependencies.copy_dependencies_local_package(dest_dir=operation_dir)
 
-        dependencies_buffer.append(f"{dest_path.relative_to(operation_dir)}")
+    local_dependencies_cmd = "\n".join(iter_dockerfile_install_command(dependencies_paths, python_major_minor))
 
-    local_dependencies_cmd = (
-        f"RUN python{python_major_minor} -m pip install --no-cache-dir " + " ".join(dependencies_buffer)
-        if len(local_dependencies) > 0
-        else ""
-    )
     return local_dependencies_cmd
 
 
-def _copy_local_code(path: Path, algo_file_path: Path, operation_dir: Path):
-    """Copy the local code given by the user to the operation directory."""
-    relative_path = path.relative_to(algo_file_path)
-    (operation_dir / relative_path.parent).mkdir(exist_ok=True)
-    if path.is_dir():
-        shutil.copytree(path, operation_dir / relative_path)
-    elif path.is_file():
-        shutil.copy(path, operation_dir / relative_path)
-    else:
-        raise ValueError(f"Does not exist {path}")
-
-
 def _create_archive(archive_path: Path, src_path: Path):
     """Create a tar archive from a folder"""
     with tarfile.open(archive_path, "w:gz") as tar:
@@ -148,6 +118,7 @@ def _get_base_docker_image(python_major_minor: str, editable_mode: bool):
                 f"Your environment uses substra-tools={substratools_image_version}. Version \
                 {MINIMAL_DOCKER_SUBSTRATOOLS_VERSION} will be used on Docker.",
                 exceptions.SubstraToolsDeprecationWarning,
+                stacklevel=2,
             )
         substratools_image_version = MINIMAL_DOCKER_SUBSTRATOOLS_VERSION
     substratools_image = _DEFAULT_SUBSTRATOOLS_IMAGE.format(
@@ -158,13 +129,13 @@ def _get_base_docker_image(python_major_minor: str, editable_mode: bool):
     return substratools_image
 
 
-def _create_substra_algo_files(
+def _create_substra_function_files(
     remote_struct: RemoteStruct,
     install_libraries: bool,
     dependencies: Dependency,
     operation_dir: Path,
 ) -> typing.Tuple[Path, Path]:
-    """Creates the necessary files from the remote struct to register the associated algorithm to substra, zip them into
+    """Creates the necessary files from the remote struct to register the associated function to substra, zip them into
         an archive (.tar.gz).
 
         Necessary files:
@@ -173,12 +144,12 @@ def _create_substra_algo_files(
             - the wheel of the current version of Substrafl and Substra
             - the Dockerfile
             - the description.md
-            - the algo.py entrypoint
+            - the function.py entrypoint
 
     Args:
         remote_struct (RemoteStruct): A representation of a substra algorithm.
         install_libraries (bool): whether we need to build the wheels and copy the files to install the libraries
-        dependencies (Dependency): Algorithm dependencies.
+        dependencies (Dependency): Function dependencies.
         operation_dir (pathlib.Path): path to the operation directory
 
         Returns:
@@ -196,10 +167,9 @@ def _create_substra_algo_files(
 
     # Build Substrafl, Substra and Substratools wheel if needed
     install_cmd = ""
-
     if install_libraries:
         # Install either from pypi wheel or repo in editable mode
-        if dependencies.editable_mode:
+        if dependencies.editable_mode or util.strtobool(os.environ.get("SUBSTRA_FORCE_EDITABLE_MODE", "False")):
             install_cmd = local_lib_wheels(
                 lib_modules=[
                     substrafl,
@@ -228,23 +198,20 @@ def _create_substra_algo_files(
     # The files to copy to the container must be in the same folder as the Dockerfile
     local_dependencies_cmd = ""
     if dependencies is not None:
-        algo_file_path = remote_struct.get_cls_file_path()
-        for path in dependencies.local_code:
-            _copy_local_code(path=path, algo_file_path=algo_file_path, operation_dir=operation_dir)
-
+        dependencies.copy_dependencies_local_code(dest_dir=operation_dir)
         if install_libraries:
             local_dep_dir = substrafl_internal / "local_dependencies"
             local_dependencies_cmd = _copy_local_packages(
                 path=local_dep_dir,
-                local_dependencies=dependencies.local_dependencies,
+                dependencies=dependencies,
                 python_major_minor=python_major_minor,
                 operation_dir=operation_dir,
             )
 
-    # Write template to algo.py
-    algo_path = operation_dir / "algo.py"
-    algo_path.write_text(
-        ALGO.format(
+    # Write template to function.py
+    function_path = operation_dir / "function.py"
+    function_path.write_text(
+        FUNCTION.format(
             substrafl_folder=SUBSTRAFL_FOLDER,
         )
     )
@@ -267,52 +234,47 @@ def _create_substra_algo_files(
             cl_deps=install_cmd,
             pypi_dependencies=pypi_dependencies_cmd,
             local_dependencies=local_dependencies_cmd,
-            # INFO: this is a temporary solution
-            # At the moment the method name is the one from the former Melody
-            # schemas, this will be variabilized soon
-            method_name=remote_struct._method_name
-            if remote_struct._remote_cls.__name__ == RemoteDataMethod.__name__
-            else "aggregate",
+            method_name=remote_struct._method_name,
         )
     )
 
     # Create necessary archive to register the operation on substra
-    archive_path = operation_dir / "algo.tar.gz"
+    archive_path = operation_dir / "function.tar.gz"
     _create_archive(archive_path=archive_path, src_path=operation_dir)
-
     return archive_path, description_path
 
 
-def register_algo(
+def register_function(
+    *,
     client: substra.Client,
     remote_struct: RemoteStruct,
     permissions: substra.sdk.schemas.Permissions,
-    inputs: typing.List[substra.sdk.schemas.AlgoInputSpec],
-    outputs: typing.List[substra.sdk.schemas.AlgoOutputSpec],
+    inputs: typing.List[substra.sdk.schemas.FunctionInputSpec],
+    outputs: typing.List[substra.sdk.schemas.FunctionOutputSpec],
     dependencies: Dependency,
 ) -> str:
-    """Automatically creates the needed files to register the composite algorithm associated to the remote_struct.
+    """Automatically creates the needed files to register the function associated to the remote_struct.
 
     Args:
         client (substra.Client): The substra client.
-        remote_struct (RemoteStruct): The substra submittable algorithm representation.
-        permissions (substra.sdk.schemas.Permissions): Permissions for the algorithm.
-        inputs (typing.List[substra.sdk.schemas.AlgoInputSpec]): List of algo inputs to be used.
-        outputs (typing.List[substra.sdk.schemas.AlgoOutputSpec]): List of algo outputs to be used.
-        dependencies (Dependency): Algorithm dependencies.
+        remote_struct (RemoteStruct): The substra submittable function representation.
+        permissions (substra.sdk.schemas.Permissions): Permissions for the function.
+        inputs (typing.List[substra.sdk.schemas.FunctionInputSpec]): List of function inputs to be used.
+        outputs (typing.List[substra.sdk.schemas.FunctionOutputSpec]): List of function outputs to be used.
+        dependencies (Dependency): Function dependencies.
 
     Returns:
-        str: Substra algorithm key.
+        str: Substra function key.
     """
     with tempfile.TemporaryDirectory(dir=str(Path.cwd().resolve()), prefix=TMP_SUBSTRAFL_PREFIX) as operation_dir:
-        archive_path, description_path = _create_substra_algo_files(
+        archive_path, description_path = _create_substra_function_files(
             remote_struct,
             dependencies=dependencies,
             install_libraries=client.backend_mode != substra.BackendType.LOCAL_SUBPROCESS,
             operation_dir=Path(operation_dir),
         )
-        key = client.add_algo(
-            substra.sdk.schemas.AlgoSpec(
+        key = client.add_function(
+            substra.sdk.schemas.FunctionSpec(
                 name=remote_struct.algo_name,
                 description=description_path,
                 file=archive_path,
@@ -325,84 +287,43 @@ def register_algo(
         return key
 
 
-def _check_metric_function(metric_function: typing.Callable):
-    """Function to check the type and the signature of a given metric function.
-
-    Args:
-        metric_function (typing.Callable): function to check.
-
-    Raises:
-        exceptions.MetricFunctionTypeError: metric_function must be of type "function"
-        exceptions.MetricFunctionSignatureError: metric_function must ONLY contains
-            datasamples and predictions_path as parameters
-    """
-
-    if not inspect.isfunction(metric_function):
-        raise exceptions.MetricFunctionTypeError("The metric_function() must be of type function.")
-
-    signature = inspect.signature(metric_function)
-    parameters = signature.parameters
-
-    if "datasamples" not in parameters:
-        raise exceptions.MetricFunctionSignatureError(
-            "The metric_function() function must contain datasamples as parameter."
-        )
-    elif "predictions_path" not in parameters:
-        raise exceptions.MetricFunctionSignatureError(
-            "The metric_function() function must contain predictions_path as parameter."
-        )
-    elif len(parameters) != 2:
-        raise exceptions.MetricFunctionSignatureError(
-            """The metric_function() function must ONLY contains datasamples and predictions_path as
-            parameters."""
-        )
-
-
-def add_metric(
+def register_metrics(
+    *,
     client: substra.Client,
-    permissions: substra.sdk.schemas.Permissions,
     dependencies: Dependency,
-    metric_function: typing.Callable,
-    metric_name: typing.Optional[str] = None,
-) -> str:
-    """Adds a metric to the Substra platform using the given metric function as the
-    algorithm to execute.
-    The metric function must be of type function, and its signature must ONLY contains
+    permissions: substra.sdk.schemas.Permissions,
+    metric_functions: typing.Dict[str, typing.Callable],
+):
+    """Adds a function to the Substra platform using the given metric functions as the
+    function to register.
+    Each metric function must be of type function, and their signature must ONLY contains
     `datasamples` and `predictions_path` as parameters. An error is raised otherwise.
 
     Args:
         client (substra.Client): The substra client.
         permissions (substra.sdk.schemas.Permissions): Permissions for the metric function.
         dependencies (Dependency): Metric function dependencies.
-        metric_function (typing.Callable): function to compute the score from the datasamples and the predictions.
-            This function is registered in substra as a metric.
-        metric_name (str, Optional): Optional name chosen by the user to identify the metric. If None,
-            the metric name is set to the 'metric_{metric_function.__name__}'.
+        metric_functions (typing.Dict[str, typing.Callable]): functions to compute the score from the datasamples and
+            the predictions. These functions are registered in substra as one function.
 
     Returns:
-        str: The metric key of the metric created from the metric function.
+        str: Substra function containing all the given metric functions.
     """
 
-    _check_metric_function(metric_function=metric_function)
-
-    class Metric:
-        def score(self, datasamples, predictions_path):
-            return metric_function(datasamples=datasamples, predictions_path=predictions_path)
-
     inputs_metrics = [
-        substra.sdk.schemas.AlgoInputSpec(
+        substra.sdk.schemas.FunctionInputSpec(
             identifier=InputIdentifiers.datasamples,
             kind=substra.sdk.schemas.AssetKind.data_sample,
             optional=False,
             multiple=True,
         ),
-        substra.sdk.schemas.AlgoInputSpec(
+        substra.sdk.schemas.FunctionInputSpec(
             identifier=InputIdentifiers.opener,
             kind=substra.sdk.schemas.AssetKind.data_manager,
             optional=False,
             multiple=False,
         ),
-        substra.sdk.schemas.AlgoInputSpec(
+        substra.sdk.schemas.FunctionInputSpec(
             identifier=InputIdentifiers.predictions,
             kind=substra.sdk.schemas.AssetKind.model,
             optional=False,
@@ -411,24 +332,34 @@ def add_metric(
     ]
 
     outputs_metrics = [
-        substra.sdk.schemas.AlgoOutputSpec(
-            identifier=OutputIdentifiers.performance,
+        substra.sdk.schemas.FunctionOutputSpec(
+            identifier=metric_function_id,
             kind=substra.sdk.schemas.AssetKind.performance,
             multiple=False,
         )
+        for metric_function_id in metric_functions
     ]
 
+    class Metric:
+        def score(self, datasamples, predictions_path, _skip=True):
+            # The _skip argument is needed to match the default signature of methods executed
+            # on substratools_methods.py.
+            return {
+                metric_function_id: metric_function(datasamples=datasamples, predictions_path=predictions_path)
+                for metric_function_id, metric_function in metric_functions.items()
+            }
+
     remote_struct = RemoteStruct(
         cls=Metric,
         cls_args=[],
         cls_kwargs={},
-        remote_cls=RemoteDataMethod,
+        remote_cls=RemoteMethod,
         method_name="score",
         method_parameters={},
-        algo_name=metric_name or "metric_" + metric_function.__name__,
+        algo_name="Evaluating",
     )
 
-    key = register_algo(
+    key = register_function(
         client=client,
         remote_struct=remote_struct,
         permissions=permissions,
