@@ -11,9 +11,9 @@ from substrafl.nodes.references.shared_state import SharedStateRef
 from substrafl.nodes.test_data_node import TestDataNode
 from substrafl.nodes.train_data_node import TrainDataNode
 from substrafl.remote import remote
-from substrafl.schemas import FedAvgAveragedState
-from substrafl.schemas import FedAvgSharedState
-from substrafl.schemas import StrategyName
+from substrafl.strategies.schemas import FedAvgAveragedState
+from substrafl.strategies.schemas import FedAvgSharedState
+from substrafl.strategies.schemas import StrategyName
 from substrafl.strategies.strategy import Strategy
 
 
@@ -45,8 +45,12 @@ class FedAvg(Strategy):
     number of samples.
     """
 
-    def __init__(self):
-        super(FedAvg, self).__init__()
+    def __init__(self, algo: Algo):
+        """
+        Args:
+            algo (Algo): The algorithm your strategy will execute (i.e. train and test on all the specified nodes)
+        """
+        super().__init__(algo=algo)
 
         # current local and share states references of the client
         self._local_states: Optional[List[LocalStateRef]] = None
@@ -63,11 +67,12 @@ class FedAvg(Strategy):
 
     def perform_round(
         self,
-        algo: Algo,
+        *,
         train_data_nodes: List[TrainDataNode],
         aggregation_node: AggregationNode,
         round_idx: int,
         clean_models: bool,
+        additional_orgs_permissions: Optional[set] = None,
     ):
         """One round of the Federated Averaging strategy consists in:
             - if ``round_idx==0``: initialize the strategy by performing a local update
@@ -77,7 +82,6 @@ class FedAvg(Strategy):
             - perform a local update (train on n mini-batches) of the models on each train data nodes
 
         Args:
-            algo (Algo): User defined algorithm: describes the model train and predict methods
             train_data_nodes (typing.List[TrainDataNode]): List of the nodes on which to perform
                 local updates.
             aggregation_node (AggregationNode): Node without data, used to perform
@@ -86,48 +90,55 @@ class FedAvg(Strategy):
             clean_models (bool): Clean the intermediary models of this round on the Substra platform.
                 Set it to False if you want to download or re-use intermediary models. This causes the disk
                 space to fill quickly so should be set to True unless needed.
+            additional_orgs_permissions (typing.Optional[set]): Additional permissions to give to the model outputs
+                after training, in order to test the model on an other organization.
         """
         if aggregation_node is None:
             raise ValueError("In FedAvg strategy aggregation node cannot be None")
 
-        if round_idx == 0:
-            # Initialization of the strategy by performing a local update on each train data organization
-            assert self._local_states is None
+        if round_idx == 1:
+            # First round of the strategy by performing a local update on each train data node
+            # We consider this step as part of the initialization and tag it as round 0.
             assert self._shared_states is None
             self._perform_local_updates(
-                algo=algo,
                 train_data_nodes=train_data_nodes,
                 current_aggregation=None,
-                round_idx=round_idx,
+                round_idx=0,
                 aggregation_id=aggregation_node.organization_id,
+                additional_orgs_permissions=additional_orgs_permissions or set(),
                 clean_models=clean_models,
             )
 
-        else:
-            current_aggregation = aggregation_node.update_states(
-                self.avg_shared_states(shared_states=self._shared_states, _algo_name="Aggregating"),  # type: ignore
-                round_idx=round_idx,
-                authorized_ids=list(set([train_data_node.organization_id for train_data_node in train_data_nodes])),
-                clean_models=clean_models,
-            )
+        current_aggregation = aggregation_node.update_states(
+            operation=self.avg_shared_states(shared_states=self._shared_states, _algo_name="Aggregating"),
+            round_idx=round_idx,
+            authorized_ids=set([train_data_node.organization_id for train_data_node in train_data_nodes]),
+            clean_models=clean_models,
+        )
 
-            self._perform_local_updates(
-                algo=algo,
-                train_data_nodes=train_data_nodes,
-                current_aggregation=current_aggregation,
-                round_idx=round_idx,
-                aggregation_id=aggregation_node.organization_id,
-                clean_models=clean_models,
-            )
+        self._perform_local_updates(
+            train_data_nodes=train_data_nodes,
+            current_aggregation=current_aggregation,
+            round_idx=round_idx,
+            aggregation_id=aggregation_node.organization_id,
+            additional_orgs_permissions=additional_orgs_permissions or set(),
+            clean_models=clean_models,
+        )
 
-    def predict(
+    def perform_predict(
         self,
-        algo: Algo,
         test_data_nodes: List[TestDataNode],
         train_data_nodes: List[TrainDataNode],
         round_idx: int,
     ):
+        """Perform prediction on test_data_nodes.
 
+        Args:
+            test_data_nodes (List[TestDataNode]): test data nodes to perform the prediction from the algo on.
+            train_data_nodes (List[TrainDataNode]): train data nodes the model has been trained
+                on.
+            round_idx (int): round index.
+        """
         for test_data_node in test_data_nodes:
             matching_train_nodes = [
                 train_data_node
@@ -135,22 +146,21 @@ class FedAvg(Strategy):
                 if train_data_node.organization_id == test_data_node.organization_id
             ]
             if len(matching_train_nodes) == 0:
-                raise NotImplementedError("Cannot test on a organization we did not train on for now.")
+                node_index = 0
+            else:
+                node_index = train_data_nodes.index(matching_train_nodes[0])
 
-            train_data_node = matching_train_nodes[0]
-            organization_index = train_data_nodes.index(train_data_node)
             assert self._local_states is not None, "Cannot predict if no training has been done beforehand."
-            local_state = self._local_states[organization_index]
+            local_state = self._local_states[node_index]
 
             test_data_node.update_states(
-                traintuple_id=local_state.key,
-                operation=algo.predict(
+                traintask_id=local_state.key,
+                operation=self.algo.predict(
                     data_samples=test_data_node.test_data_sample_keys,
-                    shared_state=None,
-                    _algo_name=f"Testing with {algo.__class__.__name__}",
+                    _algo_name=f"Predicting with {self.algo.__class__.__name__}",
                 ),
                 round_idx=round_idx,
-            )  # Init state for testtuple
+            )  # Init state for testtask
 
     @remote
     def avg_shared_states(self, shared_states: List[FedAvgSharedState]) -> FedAvgAveragedState:
@@ -188,44 +198,40 @@ class FedAvg(Strategy):
                 "Your shared_states is empty. Please ensure that "
                 "the train method of your algorithm returns a FedAvgSharedState object."
             )
-
+        parameters_update_len = len(shared_states[0].parameters_update)
         assert all(
-            [
-                len(shared_state.parameters_update) == len(shared_states[0].parameters_update)
-                for shared_state in shared_states
-            ]
+            [len(shared_state.parameters_update) == parameters_update_len for shared_state in shared_states]
         ), "Not the same number of layers for every input parameters."
 
         n_all_samples = sum([state.n_samples for state in shared_states])
 
-        averaged_states = list()
-        for idx in range(len(shared_states[0].parameters_update)):
-            states = list()
-            for state in shared_states:
-                states.append(state.parameters_update[idx] * (state.n_samples / n_all_samples))
+        averaged_states = []
+        for idx in range(parameters_update_len):
+            states = [state.parameters_update[idx] * (state.n_samples / n_all_samples) for state in shared_states]
             averaged_states.append(np.sum(states, axis=0))
 
         return FedAvgAveragedState(avg_parameters_update=averaged_states)
 
     def _perform_local_updates(
         self,
-        algo: Algo,
         train_data_nodes: List[TrainDataNode],
         current_aggregation: Optional[SharedStateRef],
         round_idx: int,
         aggregation_id: str,
+        additional_orgs_permissions: set,
         clean_models: bool,
     ):
         """Perform a local update (train on n mini-batches) of the models
         on each train data nodes.
 
         Args:
-            algo (Algo): User defined algorithm: describes the model train and predict methods
             train_data_nodes (typing.List[TrainDataNode]): List of the organizations on which to perform
             local updates current_aggregation (SharedStateRef, Optional): Reference of an aggregation operation to
                 be passed as input to each local training
             round_idx (int): Round number, it starts at 1.
             aggregation_id (str): Id of the aggregation node the shared state is given to.
+            additional_orgs_permissions (set): Additional permissions to give to the model outputs
+                after training, in order to test the model on an other organization.
             clean_models (bool): Clean the intermediary models of this round on the Substra platform.
                 Set it to False if you want to download or re-use intermediary models. This causes the disk
                 space to fill quickly so should be set to True unless needed.
@@ -235,17 +241,18 @@ class FedAvg(Strategy):
         next_shared_states = []
 
         for i, node in enumerate(train_data_nodes):
-            # define composite tuples (do not submit yet)
-            # for each composite tuple give description of Algo instead of a key for an algo
+            # define train tasks (do not submit yet)
+            # for each train task give description of Algo instead of a key for an algo
             next_local_state, next_shared_state = node.update_states(
-                algo.train(  # type: ignore
+                operation=self.algo.train(
                     node.data_sample_keys,
                     shared_state=current_aggregation,
-                    _algo_name=f"Training with {algo.__class__.__name__}",
+                    _algo_name=f"Training with {self.algo.__class__.__name__}",
                 ),
                 local_state=self._local_states[i] if self._local_states is not None else None,
                 round_idx=round_idx,
-                authorized_ids=list(set([node.organization_id, aggregation_id])),
+                authorized_ids=set([node.organization_id]) | additional_orgs_permissions,
+                aggregation_id=aggregation_id,
                 clean_models=clean_models,
             )
             # keep the states in a list: one/organization
